@@ -56,11 +56,13 @@ pub mod xdp;
 
 pub use bpf_sys::uname;
 use bpf_sys::{
-    bpf_attach_type_BPF_SK_SKB_STREAM_PARSER, bpf_attach_type_BPF_SK_SKB_STREAM_VERDICT,
-    bpf_attach_type_BPF_TRACE_ITER, bpf_create_map_attr, bpf_create_map_xattr, bpf_insn,
-    bpf_iter_create, bpf_link_create, bpf_load_program_xattr, bpf_map_def, bpf_map_info,
-    bpf_map_type_BPF_MAP_TYPE_ARRAY, bpf_map_type_BPF_MAP_TYPE_PERCPU_ARRAY, bpf_prog_type,
-    BPF_ANY,
+    bpf_attach_type_BPF_SK_LOOKUP, bpf_attach_type_BPF_SK_SKB_STREAM_PARSER,
+    bpf_attach_type_BPF_SK_SKB_STREAM_VERDICT, bpf_attach_type_BPF_TRACE_ITER, bpf_create_map_attr,
+    bpf_create_map_xattr, bpf_insn, bpf_iter_create, bpf_link_create, bpf_load_program_xattr,
+    bpf_map_def, bpf_map_info, bpf_map_type_BPF_MAP_TYPE_ARRAY, bpf_map_type_BPF_MAP_TYPE_HASH,
+    bpf_map_type_BPF_MAP_TYPE_LRU_HASH, bpf_map_type_BPF_MAP_TYPE_LRU_PERCPU_HASH,
+    bpf_map_type_BPF_MAP_TYPE_PERCPU_ARRAY, bpf_map_type_BPF_MAP_TYPE_PERCPU_HASH,
+    bpf_map_type_BPF_MAP_TYPE_PERF_EVENT_ARRAY, bpf_prog_type, BPF_ANY,
 };
 use goblin::elf::{reloc::RelocSection, section_header as hdr, Elf, SectionHeader, Sym};
 
@@ -150,6 +152,7 @@ pub enum Program {
     StreamParser(StreamParser),
     StreamVerdict(StreamVerdict),
     TaskIter(TaskIter),
+    SkLookup(SkLookup),
 }
 
 struct ProgramData {
@@ -236,6 +239,71 @@ pub struct TaskIter {
     link_fd: Option<RawFd>,
 }
 
+/// Type to work with [`sk_lookup`] BPF programs.
+///
+/// `sk_lookup` programs were introduced with Linux 5.9 and make it possible to
+/// programmatically perform socket lookup for new connections.
+/// This can be used, for instance, to listen on a large number of addresses and ports
+/// with a single socket.
+///
+/// In order to take effect, `sk_lookup` programs must be attached to a
+/// network namespace, which can be done with the [`attach_sk_lookup`] method.
+///
+/// # Example
+///
+/// The userland code for listening on a port range could look something like this.
+///
+/// ```no_run
+/// # static SK_LOOKUP: &[u8] = &[];
+/// use std::net::TcpListener;
+/// use std::os::unix::io::AsRawFd
+///
+/// use redbpf::{HashMap, SockMap};
+/// use redbpf::load::Loader;
+///
+/// let mut listener = TcpListener::bind(("127.0.0.1", 12345)).unwrap();
+/// let mut loaded = Loader::load(SK_LOOKUP).unwrap();
+///
+/// // Pass the listener fd to the BPF program
+/// let mut socket = SockMap::new(loaded.map("socket")).unwrap();
+/// socket.set(0, listener.as_raw_fd());
+///
+/// // Pass our port range to the BPF program
+/// let mut ports = HashMap::<u16, u8>::new(loaded.map("ports")).unwrap();
+/// for port in 80..430 {
+///     ports.set(port, 1);
+/// }
+///
+/// // Attach the BPF program to the current process' network namespace
+/// loaded
+///     .sk_lookup_mut("range_listener")
+///     .unwrap()
+///     .attach_sk_lookup("/proc/self/ns/net")
+///     .unwrap();
+///
+/// loop {
+///     let (client, _) = listener.accept().unwrap();
+///     let addr = client.local_addr().unwrap();
+///     println!("accepted new connection on `{}`", addr);
+/// }
+/// ```
+///
+/// [`sk_lookup`]: https://github.com/torvalds/linux/blob/master/Documentation/bpf/prog_sk_lookup.rst
+pub struct SkLookup {
+    common: ProgramData,
+    link: Option<(RawFd, RawFd)>,
+}
+
+/// A base BPF map data structure
+///
+/// It is a base data structure that contains a map definition and auxiliary
+/// data. It just hods data but it does not provide any useful API to users.
+/// See [`HashMap`](./struct.HashMap.html),
+/// [`LruHashMap`](./struct.LruHashMap.html),
+/// [`PerCpuHashMap`](./struct.PerCpuHashMap.html),
+/// [`LruPerCpuHashMap`](./struct.LruPerCpuHashMap.html),
+/// [`Array`](./struct.Array.html), [`PerCpuArray`](./struct.PerCpuArray.html)
+/// that wrap `Map` to provide API of BPF maps to userspace programs.
 #[derive(Debug)]
 pub struct Map {
     pub name: String,
@@ -259,12 +327,51 @@ enum MapBuilder<'a> {
     ExistingMap(Map),
 }
 
+/// A BPF hash map structure
+///
+/// This provides higher level API for BPF maps whose type is
+/// `BPF_MAP_TYPE_HASH`
 pub struct HashMap<'a, K: Clone, V: Clone> {
     base: &'a Map,
     _k: PhantomData<K>,
     _v: PhantomData<V>,
 }
 
+/// A BPF LRU hash map structure
+///
+/// This provides higher level API for BPF maps whose type is
+/// `BPF_MAP_TYPE_LRU_HASH`
+pub struct LruHashMap<'a, K: Clone, V: Clone> {
+    base: &'a Map,
+    _k: PhantomData<K>,
+    _v: PhantomData<V>,
+}
+
+/// A per-cpu BPF hash map structure
+///
+/// This provides higher level API for BPF maps whose type is
+/// `BPF_MAP_TYPE_PERCPU_HASH`
+pub struct PerCpuHashMap<'a, K: Clone, V: Clone> {
+    base: &'a Map,
+    _k: PhantomData<K>,
+    _v: PhantomData<PerCpuValues<V>>,
+}
+
+/// An LRU per-cpu BPF hash map structure
+///
+/// This provides higher level API for BPF maps whose type is
+/// `BPF_MAP_TYPE_LRU_PERCPU_HASH`
+pub struct LruPerCpuHashMap<'a, K: Clone, V: Clone> {
+    base: &'a Map,
+    _k: PhantomData<K>,
+    _v: PhantomData<PerCpuValues<V>>,
+}
+
+/// A stacktrace BPF map structure
+///
+/// Stacktrace map provides a feature of getting an array of instruction
+/// pointers that are stored in the BPF map whose type is
+/// `BPF_MAP_TYPE_STACK_TRACE`.
 pub struct StackTrace<'a> {
     base: &'a Map,
 }
@@ -343,6 +450,11 @@ pub struct RelocationInfo {
     sym_idx: usize,
 }
 
+trait MapIterable<K: Clone, V: Clone> {
+    fn get(&self, key: K) -> Option<V>;
+    fn next_key(&self, key: Option<K>) -> Option<K>;
+}
+
 impl Program {
     #[allow(clippy::unnecessary_wraps)]
     fn new(kind: &str, name: &str, code: &[u8]) -> Result<Program> {
@@ -384,6 +496,7 @@ impl Program {
             }),
             "streamparser" => Program::StreamParser(StreamParser { common }),
             "streamverdict" => Program::StreamVerdict(StreamVerdict { common }),
+            "sk_lookup" => Program::SkLookup(SkLookup { common, link: None }),
             _ => return Err(Error::Section(kind.to_string())),
         })
     }
@@ -426,6 +539,7 @@ impl Program {
             TracePoint(_) => bpf_sys::bpf_prog_type_BPF_PROG_TYPE_TRACEPOINT,
             StreamParser(_) | StreamVerdict(_) => bpf_sys::bpf_prog_type_BPF_PROG_TYPE_SK_SKB,
             TaskIter(_) => bpf_sys::bpf_prog_type_BPF_PROG_TYPE_TRACING,
+            SkLookup(_) => bpf_sys::bpf_prog_type_BPF_PROG_TYPE_SK_LOOKUP,
         }
     }
 
@@ -441,6 +555,7 @@ impl Program {
             StreamParser(p) => &p.common,
             StreamVerdict(p) => &p.common,
             TaskIter(p) => &p.common,
+            SkLookup(p) => &p.common,
         }
     }
 
@@ -456,6 +571,7 @@ impl Program {
             StreamParser(p) => &mut p.common,
             StreamVerdict(p) => &mut p.common,
             TaskIter(p) => &mut p.common,
+            SkLookup(p) => &mut p.common,
         }
     }
 
@@ -499,12 +615,19 @@ impl Program {
         attr.license = clicense.as_ptr();
         attr.log_level = 0;
 
-        if let Program::TaskIter(bpf_iter) = self {
-            attr.expected_attach_type = bpf_attach_type_BPF_TRACE_ITER;
-            attr.__bindgen_anon_2.attach_btf_id = bpf_iter.attach_btf_id;
-        } else {
-            attr.expected_attach_type = 0;
-            attr.__bindgen_anon_1.kern_version = kernel_version;
+        match self {
+            Program::TaskIter(bpf_iter) => {
+                attr.expected_attach_type = bpf_attach_type_BPF_TRACE_ITER;
+                attr.__bindgen_anon_2.attach_btf_id = bpf_iter.attach_btf_id;
+            }
+            Program::SkLookup(_) => {
+                attr.expected_attach_type = bpf_attach_type_BPF_SK_LOOKUP;
+                attr.__bindgen_anon_1.kern_version = kernel_version;
+            }
+            _ => {
+                attr.expected_attach_type = 0;
+                attr.__bindgen_anon_1.kern_version = kernel_version;
+            }
         }
 
         // do not pass log buffer. it is filled with verifier's log but
@@ -879,6 +1002,40 @@ impl XDP {
         }
     }
 
+    /// Detach the XDP program.
+    ///
+    /// Detach the XDP program from the given network interface, if attached.
+    ///
+    /// # Example
+    /// ```no_run
+    /// # use redbpf::{Module, xdp};
+    /// # let mut module = Module::parse(&std::fs::read("file.elf").unwrap()).unwrap();
+    /// # for uprobe in module.xdps_mut() {
+    /// uprobe.attach_xdp("eth0", xdp::Flags::default()).unwrap();
+    /// uprobe.detach_xdp("eth0").unwrap();
+    /// # }
+    /// ```
+    pub fn detach_xdp(&mut self, interface: &str) -> Result<()> {
+        // The linear search here isn't great, but self.interfaces will almost always be short.
+        let index = self
+            .interfaces
+            .iter()
+            .enumerate()
+            .find_map(|(i, v)| (v.as_str() == interface).then(|| i))
+            .ok_or(Error::ProgramNotLoaded)?;
+        if let Err(e) = unsafe { detach_xdp(interface) } {
+            if let Error::IO(ref oserr) = e {
+                error!(
+                    "error detaching xdp from interface {}: {}",
+                    interface, oserr
+                );
+            }
+            return Err(e);
+        }
+        self.interfaces.swap_remove(index);
+        Ok(())
+    }
+
     pub fn name(&self) -> String {
         self.common.name.to_string()
     }
@@ -887,7 +1044,7 @@ impl XDP {
 impl Drop for XDP {
     fn drop(&mut self) {
         for interface in self.interfaces.iter() {
-            let _ = unsafe { attach_xdp(interface, -1, 0) };
+            let _ = unsafe { detach_xdp(interface) };
         }
     }
 }
@@ -943,6 +1100,10 @@ unsafe fn attach_xdp(dev_name: &str, progfd: libc::c_int, flags: libc::c_uint) -
     Ok(())
 }
 
+unsafe fn detach_xdp(dev_name: &str) -> Result<()> {
+    attach_xdp(dev_name, -1, 0)
+}
+
 impl SocketFilter {
     /// Attach the socket filter program.
     ///
@@ -981,6 +1142,47 @@ impl SocketFilter {
     }
 }
 
+impl SkLookup {
+    /// Attach the `sk_lookup` to the given network namespace.
+    ///
+    /// In most cases it should be attached to `/proc/self/ns/net`.
+    pub fn attach_sk_lookup(&mut self, namespace: &str) -> Result<()> {
+        if self.link.is_some() {
+            return Err(Error::ProgramAlreadyLinked);
+        }
+
+        let fd = self.common.fd.ok_or(Error::ProgramNotLoaded)?;
+        unsafe {
+            let namespace = CString::new(namespace)?;
+            let nfd = libc::open(namespace.as_ptr(), libc::O_RDONLY | libc::O_CLOEXEC);
+            if nfd < 0 {
+                return Err(Error::IO(io::Error::last_os_error()));
+            }
+
+            let lfd = bpf_link_create(fd, nfd, bpf_attach_type_BPF_SK_LOOKUP, ptr::null());
+            if lfd < 0 {
+                libc::close(nfd);
+                return Err(Error::IO(io::Error::last_os_error()));
+            }
+
+            self.link = Some((nfd, lfd));
+        }
+
+        Ok(())
+    }
+}
+
+impl Drop for SkLookup {
+    fn drop(&mut self) {
+        if let Some((nfd, lfd)) = self.link.take() {
+            unsafe {
+                libc::close(lfd);
+                libc::close(nfd);
+            }
+        }
+    }
+}
+
 impl Module {
     pub fn parse(bytes: &[u8]) -> Result<Module> {
         ModuleBuilder::parse(bytes)?.to_module()
@@ -1142,6 +1344,26 @@ impl Module {
         self.stream_verdicts_mut().find(|p| p.common.name == name)
     }
 
+    pub fn sk_lookups(&self) -> impl Iterator<Item = &SkLookup> {
+        use Program::*;
+        self.programs.iter().filter_map(|prog| match prog {
+            SkLookup(p) => Some(p),
+            _ => None,
+        })
+    }
+
+    pub fn sk_lookups_mut(&mut self) -> impl Iterator<Item = &mut SkLookup> {
+        use Program::*;
+        self.programs.iter_mut().filter_map(|prog| match prog {
+            SkLookup(p) => Some(p),
+            _ => None,
+        })
+    }
+
+    pub fn sk_lookup_mut(&mut self, name: &str) -> Option<&mut SkLookup> {
+        self.sk_lookups_mut().find(|p| p.common.name == name)
+    }
+
     pub fn task_iters(&self) -> impl Iterator<Item = &TaskIter> {
         use Program::*;
         self.programs.iter().filter_map(|prog| match prog {
@@ -1265,7 +1487,8 @@ impl<'a> ModuleBuilder<'a> {
                 | (hdr::SHT_PROGBITS, Some(kind @ "xdp"), Some(name))
                 | (hdr::SHT_PROGBITS, Some(kind @ "socketfilter"), Some(name))
                 | (hdr::SHT_PROGBITS, Some(kind @ "streamparser"), Some(name))
-                | (hdr::SHT_PROGBITS, Some(kind @ "streamverdict"), Some(name)) => {
+                | (hdr::SHT_PROGBITS, Some(kind @ "streamverdict"), Some(name))
+                | (hdr::SHT_PROGBITS, Some(kind @ "sk_lookup"), Some(name)) => {
                     let prog = Program::new(kind, name, &content)?;
                     programs.insert(shndx, prog);
                 }
@@ -1333,7 +1556,11 @@ impl<'a> ModuleBuilder<'a> {
             if self.programs.contains_key(&rel.target_sec_idx) {
                 if let Err(_) = rel.apply(&mut self.programs, &maps, &symtab) {
                     // means that not normal case, we should rely on symbol value instead of section header index
-                    rel.apply_with_symmap(&mut self.programs, &symval_to_maps, &symtab)?;
+                    rel.apply_with_symmap(&mut self.programs, &symval_to_maps, &symtab)
+                        .map_err(|e| {
+                            error!("can not relocate map");
+                            e
+                        })?;
                 }
             }
         }
@@ -1754,9 +1981,11 @@ impl<'base, K: Clone, V: Clone> HashMap<'base, K, V> {
     pub fn new(base: &Map) -> Result<HashMap<K, V>> {
         if mem::size_of::<K>() != base.config.key_size as usize
             || mem::size_of::<V>() != base.config.value_size as usize
+            || (bpf_map_type_BPF_MAP_TYPE_HASH != base.config.type_
+                && bpf_map_type_BPF_MAP_TYPE_PERF_EVENT_ARRAY != base.config.type_)
         {
             error!(
-                "map definitions (sizes of key and value) of base `Map' and
+                "map definitions (map type and key/value size) of base `Map' and
             `HashMap' do not match"
             );
             return Err(Error::Map);
@@ -1769,46 +1998,211 @@ impl<'base, K: Clone, V: Clone> HashMap<'base, K, V> {
         })
     }
 
-    pub fn set(&self, mut key: K, mut value: V) {
-        unsafe {
-            bpf_sys::bpf_map_update_elem(
-                self.base.fd,
-                &mut key as *mut _ as *mut _,
-                &mut value as *mut _ as *mut _,
-                0,
-            );
-        }
+    pub fn set(&self, key: K, value: V) {
+        let _ = bpf_map_set(self.base.fd, key, value);
     }
 
-    pub fn get(&self, mut key: K) -> Option<V> {
-        let mut value = MaybeUninit::zeroed();
-        if unsafe {
-            bpf_sys::bpf_map_lookup_elem(
-                self.base.fd,
-                &mut key as *mut _ as *mut _,
-                &mut value as *mut _ as *mut _,
-            )
-        } < 0
-        {
-            return None;
-        }
-        Some(unsafe { value.assume_init() })
+    pub fn get(&self, key: K) -> Option<V> {
+        bpf_map_get(self.base.fd, key)
     }
 
-    pub fn delete(&self, mut key: K) {
-        unsafe {
-            bpf_sys::bpf_map_delete_elem(self.base.fd, &mut key as *mut _ as *mut _);
-        }
+    pub fn delete(&self, key: K) {
+        let _ = bpf_map_delete(self.base.fd, key);
     }
 
-    pub fn iter<'a>(&'a self) -> MapIter<'a, '_, K, V> {
+    /// Return an iterator over all items in the map
+    pub fn iter<'a>(&'a self) -> MapIter<'a, K, V> {
         MapIter {
-            map: self,
-            key: None,
+            iterable: self,
+            last_key: None,
         }
     }
 }
 
+impl<K: Clone, V: Clone> MapIterable<K, V> for HashMap<'_, K, V> {
+    fn get(&self, key: K) -> Option<V> {
+        HashMap::get(self, key)
+    }
+
+    fn next_key(&self, key: Option<K>) -> Option<K> {
+        bpf_map_get_next_key(self.base.fd, key)
+    }
+}
+
+impl<'base, K: Clone, V: Clone> LruHashMap<'base, K, V> {
+    pub fn new(base: &Map) -> Result<LruHashMap<K, V>> {
+        if mem::size_of::<K>() != base.config.key_size as usize
+            || mem::size_of::<V>() != base.config.value_size as usize
+            || bpf_map_type_BPF_MAP_TYPE_LRU_HASH != base.config.type_
+        {
+            error!(
+                "map definitions (map type and key/value sizes) of base `Map' and `LruHashMap' do not match"
+            );
+            return Err(Error::Map);
+        }
+
+        Ok(LruHashMap {
+            base,
+            _k: PhantomData,
+            _v: PhantomData,
+        })
+    }
+
+    pub fn set(&self, key: K, value: V) {
+        let _ = bpf_map_set(self.base.fd, key, value);
+    }
+
+    pub fn get(&self, key: K) -> Option<V> {
+        bpf_map_get(self.base.fd, key)
+    }
+
+    pub fn delete(&self, key: K) {
+        let _ = bpf_map_delete(self.base.fd, key);
+    }
+
+    /// Return an iterator over all items in the map
+    pub fn iter<'a>(&'a self) -> MapIter<'a, K, V> {
+        MapIter {
+            iterable: self,
+            last_key: None,
+        }
+    }
+}
+
+impl<K: Clone, V: Clone> MapIterable<K, V> for LruHashMap<'_, K, V> {
+    fn get(&self, key: K) -> Option<V> {
+        LruHashMap::<'_, K, V>::get(self, key)
+    }
+
+    fn next_key(&self, key: Option<K>) -> Option<K> {
+        bpf_map_get_next_key(self.base.fd, key)
+    }
+}
+
+impl<'base, K: Clone, V: Clone> PerCpuHashMap<'base, K, V> {
+    pub fn new(base: &Map) -> Result<PerCpuHashMap<K, V>> {
+        if mem::size_of::<K>() != base.config.key_size as usize
+            || mem::size_of::<V>() != base.config.value_size as usize
+            || bpf_map_type_BPF_MAP_TYPE_PERCPU_HASH != base.config.type_
+        {
+            error!("map definitions (size of key/value and map type) of base `Map' and `PerCpuHashMap' do not match");
+            return Err(Error::Map);
+        }
+
+        Ok(PerCpuHashMap {
+            base,
+            _k: PhantomData,
+            _v: PhantomData,
+        })
+    }
+
+    /// Set per-cpu `values` to the BPF map at `key`
+    ///
+    /// The number of elements in `values` should be equal to the number of
+    /// possible CPUs. This requirement is automatically fulfilled when
+    /// `values` is created by
+    /// [`PerCpuValues::new`](./struct.PerCpuValues.html#method.new)
+    ///
+    /// `Err` can be returned if the number of elements is wrong or underlying
+    /// bpf_map_update_elem function returns a negative value.
+    pub fn set(&self, key: K, values: PerCpuValues<V>) -> Result<()> {
+        bpf_percpu_map_set(self.base.fd, key, values)
+    }
+
+    /// Get per-cpu values corresponding to the `key` from the BPF map
+    ///
+    /// If `key` is found, `Some([PerCpuValues](./struct.PerCpuValues.html))`
+    /// is returned.
+    pub fn get(&self, key: K) -> Option<PerCpuValues<V>> {
+        bpf_percpu_map_get(self.base.fd, key)
+    }
+
+    /// Delete `key` from the BPF map
+    pub fn delete(&self, key: K) {
+        let _ = bpf_map_delete(self.base.fd, key);
+    }
+
+    /// Return an iterator over all items in the map
+    pub fn iter<'a>(&'a self) -> MapIter<'a, K, PerCpuValues<V>> {
+        MapIter {
+            iterable: self,
+            last_key: None,
+        }
+    }
+}
+
+impl<K: Clone, V: Clone> MapIterable<K, PerCpuValues<V>> for PerCpuHashMap<'_, K, V> {
+    fn get(&self, key: K) -> Option<PerCpuValues<V>> {
+        PerCpuHashMap::get(self, key)
+    }
+
+    fn next_key(&self, key: Option<K>) -> Option<K> {
+        bpf_map_get_next_key(self.base.fd, key)
+    }
+}
+
+impl<'base, K: Clone, V: Clone> LruPerCpuHashMap<'base, K, V> {
+    pub fn new(base: &Map) -> Result<LruPerCpuHashMap<K, V>> {
+        if mem::size_of::<K>() != base.config.key_size as usize
+            || mem::size_of::<V>() != base.config.value_size as usize
+            || bpf_map_type_BPF_MAP_TYPE_LRU_PERCPU_HASH != base.config.type_
+        {
+            error!("map definitions (size of key/value and map type) of base `Map' and `LruPerCpuHashMap' do not match");
+            return Err(Error::Map);
+        }
+
+        Ok(LruPerCpuHashMap {
+            base,
+            _k: PhantomData,
+            _v: PhantomData,
+        })
+    }
+
+    /// Set per-cpu `values` to the BPF map at `key`
+    ///
+    /// The number of elements in `values` should be equal to the number of
+    /// possible CPUs. This requirement is automatically fulfilled when
+    /// `values` is created by
+    /// [`PerCpuValues::new`](./struct.PerCpuValues.html#method.new)
+    ///
+    /// `Err` can be returned if the number of elements is wrong or underlying
+    /// bpf_map_update_elem function returns a negative value.
+    pub fn set(&self, key: K, values: PerCpuValues<V>) -> Result<()> {
+        bpf_percpu_map_set(self.base.fd, key, values)
+    }
+
+    /// Get per-cpu values corresponding to the `key` from the BPF map
+    ///
+    /// If `key` is found, `Some([PerCpuValues](./struct.PerCpuValues.html))`
+    /// is returned.
+    pub fn get(&self, key: K) -> Option<PerCpuValues<V>> {
+        bpf_percpu_map_get(self.base.fd, key)
+    }
+
+    /// Delete `key` from the BPF map
+    pub fn delete(&self, key: K) {
+        let _ = bpf_map_delete(self.base.fd, key);
+    }
+
+    /// Return an iterator over all items in the map
+    pub fn iter<'a>(&'a self) -> MapIter<'a, K, PerCpuValues<V>> {
+        MapIter {
+            iterable: self,
+            last_key: None,
+        }
+    }
+}
+
+impl<K: Clone, V: Clone> MapIterable<K, PerCpuValues<V>> for LruPerCpuHashMap<'_, K, V> {
+    fn get(&self, key: K) -> Option<PerCpuValues<V>> {
+        LruPerCpuHashMap::get(self, key)
+    }
+
+    fn next_key(&self, key: Option<K>) -> Option<K> {
+        bpf_map_get_next_key(self.base.fd, key)
+    }
+}
+
 impl<'base, T: Clone> Array<'base, T> {
     /// Create `Array` map from `base`
     pub fn new(base: &Map) -> Result<Array<T>> {
@@ -1886,12 +2280,14 @@ fn round_up<T>(unit_size: usize) -> usize {
 /// the same with [`cpus::get_possible_num`](./cpus/fn.get_possible_num.html).
 /// It also implements `Deref` and `DerefMut` so it can be used as a normal
 /// array.
+///
 /// # Example
 /// ```no_run
 /// use redbpf::PerCpuValues;
 /// let mut values = PerCpuValues::<u64>::new(0);
 /// values[0] = 1;
 /// ```
+#[derive(Clone, Debug)]
 pub struct PerCpuValues<T: Clone>(Box<[T]>);
 
 impl<T: Clone> PerCpuValues<T> {
@@ -1902,12 +2298,19 @@ impl<T: Clone> PerCpuValues<T> {
     pub fn new(default_value: T) -> Self {
         let count = cpus::get_possible_num();
         let v = vec![default_value; count];
-        Self(v.into_boxed_slice())
+        Self(v.into())
     }
+}
 
-    // This is called by `get` methods of per-cpu map structures
-    fn from_boxed_slice(v: Box<[T]>) -> Self {
-        Self(v)
+impl<T: Clone> From<Box<[T]>> for PerCpuValues<T> {
+    fn from(values: Box<[T]>) -> Self {
+        Self(values)
+    }
+}
+
+impl<T: Clone> From<Vec<T>> for PerCpuValues<T> {
+    fn from(values: Vec<T>) -> Self {
+        Self::from(values.into_boxed_slice())
     }
 }
 
@@ -2016,7 +2419,7 @@ impl<'base, T: Clone> PerCpuArray<'base, T> {
             }
         }
 
-        Some(PerCpuValues::from_boxed_slice(values.into_boxed_slice()))
+        Some(values.into())
     }
 
     /// Get length of array map
@@ -2091,51 +2494,21 @@ impl<'base> ProgramArray<'base> {
     }
 }
 
-pub struct MapIter<'a, 'b, K: Clone, V: Clone> {
-    map: &'a HashMap<'b, K, V>,
-    key: Option<K>,
+pub struct MapIter<'a, K: Clone, V: Clone> {
+    iterable: &'a dyn MapIterable<K, V>,
+    last_key: Option<K>,
 }
 
-impl<K: Clone, V: Clone> Iterator for MapIter<'_, '_, K, V> {
+impl<K: Clone, V: Clone> Iterator for MapIter<'_, K, V> {
     type Item = (K, V);
 
     fn next(&mut self) -> Option<Self::Item> {
-        let key = self.key.take();
-        self.key = match key {
-            Some(mut key) => {
-                let mut next_key = MaybeUninit::<K>::zeroed();
-                let ret = unsafe {
-                    bpf_sys::bpf_map_get_next_key(
-                        self.map.base.fd,
-                        &mut key as *mut _ as *mut _,
-                        &mut next_key as *mut _ as *mut _,
-                    )
-                };
-                if ret < 0 {
-                    None
-                } else {
-                    Some(unsafe { next_key.assume_init() })
-                }
-            }
-            None => {
-                let mut key = MaybeUninit::<K>::zeroed();
-                if unsafe {
-                    bpf_sys::bpf_map_get_next_key(
-                        self.map.base.fd,
-                        ptr::null(),
-                        &mut key as *mut _ as *mut _,
-                    )
-                } < 0
-                {
-                    None
-                } else {
-                    Some(unsafe { key.assume_init() })
-                }
-            }
-        };
-
-        let key = self.key.as_ref()?.clone();
-        Some((key.clone(), self.map.get(key).unwrap()))
+        let key = self.last_key.take();
+        self.last_key = self.iterable.next_key(key);
+        Some((
+            self.last_key.as_ref()?.clone(),
+            self.iterable.get(self.last_key.as_ref()?.clone())?,
+        ))
     }
 }
 
@@ -2396,3 +2769,130 @@ fn data<'d>(bytes: &'d [u8], shdr: &SectionHeader) -> &'d [u8] {
 
     &bytes[offset..end]
 }
+
+fn bpf_map_set<K: Clone, V: Clone>(fd: RawFd, mut key: K, mut value: V) -> Result<()> {
+    if unsafe {
+        bpf_sys::bpf_map_update_elem(
+            fd,
+            &mut key as *mut _ as *mut _,
+            &mut value as *mut _ as *mut _,
+            0,
+        )
+    } < 0
+    {
+        Err(Error::Map)
+    } else {
+        Ok(())
+    }
+}
+
+fn bpf_map_get<K: Clone, V: Clone>(fd: RawFd, mut key: K) -> Option<V> {
+    let mut value = MaybeUninit::zeroed();
+    if unsafe {
+        bpf_sys::bpf_map_lookup_elem(
+            fd,
+            &mut key as *mut _ as *mut _,
+            &mut value as *mut _ as *mut _,
+        )
+    } < 0
+    {
+        return None;
+    }
+    Some(unsafe { value.assume_init() })
+}
+
+fn bpf_map_delete<K: Clone>(fd: RawFd, mut key: K) -> Result<()> {
+    if unsafe { bpf_sys::bpf_map_delete_elem(fd, &mut key as *mut _ as *mut _) } < 0 {
+        Err(Error::Map)
+    } else {
+        Ok(())
+    }
+}
+
+fn bpf_map_get_next_key<K: Clone>(fd: RawFd, key: Option<K>) -> Option<K> {
+    if let Some(mut key) = key {
+        let mut next_key = MaybeUninit::<K>::zeroed();
+        let ret = unsafe {
+            bpf_sys::bpf_map_get_next_key(
+                fd,
+                &mut key as *mut _ as *mut _,
+                &mut next_key as *mut _ as *mut _,
+            )
+        };
+        if ret < 0 {
+            None
+        } else {
+            Some(unsafe { next_key.assume_init() })
+        }
+    } else {
+        let mut key = MaybeUninit::<K>::zeroed();
+        if unsafe { bpf_sys::bpf_map_get_next_key(fd, ptr::null(), &mut key as *mut _ as *mut _) }
+            < 0
+        {
+            None
+        } else {
+            Some(unsafe { key.assume_init() })
+        }
+    }
+}
+
+fn bpf_percpu_map_set<K: Clone, V: Clone>(
+    fd: RawFd,
+    mut key: K,
+    values: PerCpuValues<V>,
+) -> Result<()> {
+    let count = cpus::get_possible_num();
+    if values.len() != count {
+        return Err(Error::Map);
+    }
+
+    // It is needed to round up the value size to 8*N bytes
+    // cf., https://elixir.bootlin.com/linux/v5.8/source/kernel/bpf/syscall.c#L1103
+    let value_size = round_up::<V>(8);
+    let alloc_size = value_size * count;
+    let mut alloc = vec![0u8; alloc_size];
+    let mut data = alloc.as_mut_ptr();
+    for i in 0..count {
+        unsafe {
+            let dst_ptr = data.add(value_size * i) as *const V as *mut V;
+            ptr::write_unaligned::<V>(dst_ptr, values[i].clone());
+        }
+    }
+    if unsafe {
+        bpf_sys::bpf_map_update_elem(
+            fd,
+            &mut key as *mut _ as *mut _,
+            &mut data as *mut _ as *mut _,
+            0,
+        )
+    } < 0
+    {
+        Err(Error::Map)
+    } else {
+        Ok(())
+    }
+}
+
+fn bpf_percpu_map_get<K: Clone, V: Clone>(fd: RawFd, mut key: K) -> Option<PerCpuValues<V>> {
+    // It is needed to round up the value size to 8*N
+    // cf., https://elixir.bootlin.com/linux/v5.8/source/kernel/bpf/syscall.c#L1035
+    let value_size = round_up::<V>(8);
+    let count = cpus::get_possible_num();
+    let alloc_size = value_size * count;
+    let mut alloc = vec![0u8; alloc_size];
+    let data = alloc.as_mut_ptr();
+    if unsafe { bpf_sys::bpf_map_lookup_elem(fd, &mut key as *mut _ as *mut _, data as *mut _) } < 0
+    {
+        return None;
+    }
+
+    let mut values = Vec::with_capacity(count);
+    for i in 0..count {
+        unsafe {
+            let elem_ptr = data.add(value_size * i) as *const V;
+            values.push(ptr::read_unaligned(elem_ptr));
+        }
+    }
+
+    Some(values.into())
+}
